#pragma once
#ifndef TRIANGLE_H
#define TRIANGLE_H

#include <QDebug>
#include <QVector>
#include "vertex.h"
#include "mesh.h"


class Triangle {
    public:
        struct Edge {
            enum Name {
                uv,
                vw,
                wu
            };

            Name name;
            unsigned int a;//index der Kante im Dreieck
            unsigned int b;//index der Kante im Dreieck
            unsigned int c;//gegenüberliegender Vertex index
        };

        struct Neighbor {
            Triangle *triangle;
            Edge edge;
        };

        struct Neighbors {
            Neighbor uv;
            Neighbor vw;
            Neighbor wu;

            Triangle::Neighbor get_neighbor(Triangle::Edge::Name name) const;
        };

        Triangle();
        Triangle(const Triangle &other);
        Triangle(const QVector<Vertex> *vertex_buffer, unsigned int u, unsigned int v, unsigned int w,
                 unsigned int sharpness_uv = 0,
                 unsigned int sharpness_vw = 0,
                 unsigned int sharpness_wu = 0);

        Vertex u() const;
        Vertex v() const;
        Vertex w() const;

        unsigned int u_idx() const;
        unsigned int v_idx() const;
        unsigned int w_idx() const;

        unsigned int get_sharpness_uv() const;
        unsigned int get_sharpness_vw() const;
        unsigned int get_sharpness_wu() const;

        bool get_shared_edge(Triangle other, Edge &edge_a, Edge &edge_b) const;

        bool hasSharedEdge(Triangle other);

        Triangle &operator=(const Triangle &other);

        // == and < both ignore the vertex buffer, they only compare indices.
        // They are implemented so that Triangle can be used as the key of a QMap.
        bool operator==(const Triangle &other) const;
        bool operator!=(const Triangle &other) const;
        bool operator<(const Triangle &other) const;

        bool next_ctrclockwise(Edge first, Edge next) const;
        bool next_clockwise(Edge first, Edge next) const;

        void set_sharpness(Triangle::Edge::Name name, unsigned int value);

    private:

        unsigned int sharpness_uv;
        unsigned int sharpness_vw;
        unsigned int sharpness_wu;

        const QVector<Vertex> *vertex_buffer_;
        unsigned int u_;
        unsigned int v_;
        unsigned int w_;

        void rotate_indices();
        bool get_shared_edge_(const Triangle &other, Edge &edge_a, Edge &edge_b) const;
};

Triangle::Edge::Name rotate_edge_name(Triangle::Edge::Name edge);
void ibToTriangles(const QVector<Vertex> *vb, const QVector<unsigned int> &ib, const QMap<Edge,unsigned int>& sharpEdges, QVector<Triangle> &triangles);
void regularIbToTriangles(const QVector<Vertex> *vb, const QVector<unsigned int> &ib, QVector<Triangle> &triangles);
void trianglesToIB(const QVector<Triangle> &triangles, QVector<unsigned int> &ib);

QDebug operator<<(QDebug d, const Triangle &triangle);

#endif
